/*
 * (C) Copyright 2011
 * Working Solar System Production
 *  
 * BootStrap Setting fuse
 */


#include <common.h>
#include <command.h>
#include <asm/io.h>
#include <asm/errno.h>
#include <asm/arch/mx25.h>
#include <asm/arch/mx25-regs.h>
#include <asm/arch/mx25_pins.h>
#include <asm/arch/iomux.h>
#include <asm/arch/gpio.h>

#ifndef writeb
# define writeb(val, addr)	\
	*(volatile u8 *)(addr) = (u8)(val)
#endif	/* writeb */

#ifndef readb
# define readb(addr)	*(volatile u8 *)(addr)
#endif	/* readb */

/****
 * debugging
 ****/
static char __conv_f[8];
static char * __convert_fuse(u8 byte)
{
	int i, j;

	for(i = 7, j = 0; j < 8; i--, j++) {
		if ((byte >> i) & 1)
			__conv_f[j] = '1';
		else
			__conv_f[j] = '0';
	}
	return __conv_f;
}

/*
 * fuse_bank is the fuse bank index. MX25 only has 3 banks. That's why they've been enumerated.
 * fuse_row_addr: represents the address of the row where we are going to blow a fuse.
 *  Each Row addresses one byte worth of data. This means one row contains 8 fuses ready
 * to be blown!
 *  Row Index 0 is Row Address 0x0
 *  Row Index 1 is Row Address 0x4
 *  Row Index 2 is Row Address 0x8
 *  Row Index 3 is Row Address 0xC
 *  ... And so on
 * fuse_bit_addr: A number from 0 to 7 that represent the fuse bit index that we are going to
 * blow or sense.
 */
typedef enum {
	ESNS_1 = 2,
	ESNS_0 = 4,
	ESNS_N = 8	// Normal Sensing
} e_sens_strength;

typedef enum {
	FUSE_BANK_0,
	FUSE_BANK_1,
	FUSE_BANK_2,
	MAX_FUSE_BANK	//MX25 only has 3 fuse banks, 0, 1 and 2.
} e_fuse_banks;

#define IMX_IIM_BASE	IIM_BASE

static void set_fuse_address(e_fuse_banks fuse_bank,
			u16 fuse_row_addr, u8 fuse_bit_addr)
{
	u8 upper_addr;
	u8 lower_addr;
	u8 fuse_index;

	/* A fuse bank contains 256 fuse rows. for a total of 2048 fuse bits.
	 * Dividing the fuse_address by 4 will give us the fuse index. */
	fuse_index = (u8)(fuse_row_addr >> 2);

	upper_addr = ((u8)fuse_bank << 3);
	upper_addr |= ((fuse_index >> 5) & 0x7);

	lower_addr = ((fuse_index & 0x1F) << 3);
	lower_addr |= (fuse_bit_addr & 0x7);

	/* write address to UA LA register */
	writeb(upper_addr, IMX_IIM_BASE + IIM_UAF);
	writeb(lower_addr, IMX_IIM_BASE + IIM_LA);
}

static int fuse_sence(u32 bank, u32 row)
{
	u8 sdat, err, stat;
	e_fuse_banks fusebox;

	if (row > 0x3ff) {
		printf("fuse_sense: invalid row offset\n");
		return -1;
	}

	/* bank no is fusebox setting bit convert */
	switch(bank) {
	case 0:	fusebox = FUSE_BANK_0;	break;
	case 1:	fusebox = FUSE_BANK_1;	break;
	case 2:	fusebox = FUSE_BANK_2;	break;
	default:
		printf("fuse_sense: invalid bank number\n");
		return -1;
	}

	/* clear status and error registers */
	writeb(3, IMX_IIM_BASE + IIM_STATM);
	writeb(0xfe, IMX_IIM_BASE + IIM_ERR);

	/* upper and lower address halves */
	set_fuse_address(fusebox, row, 0);

	/* start fuse sensing */
	writeb(ESNS_N, IMX_IIM_BASE + IIM_FCTL);

	/* wait for sense done */
	while ((readb(IMX_IIM_BASE + IIM_STAT) & 0x80) != 0)
		;

	stat =  readb(IMX_IIM_BASE + IIM_STAT);
	writeb(stat, IMX_IIM_BASE + IIM_STAT);

	err = readb(IMX_IIM_BASE + IIM_ERR);
	if (err) {
		printf("fuse_sense: sense error (0x%02x)\n", err);
		return -1;
	}

	sdat = readb(IMX_IIM_BASE + IIM_SDAT);
	return sdat;
}

static int fuse_blow_value(u32 bank, u32 row, u32 value)
{
	int bit, ret = 0;
	u8 err, stat;

	if ((bank > 7) || (row > 0x3ff) || (value > 0xff)) {
		printf("fuse_blow: invalid parameter\n");
		return 1;
	}

	/* clear status and error registers */
	writeb(3, IMX_IIM_BASE + IIM_STATM);
	writeb(0xfe, IMX_IIM_BASE + IIM_ERR);

	/* unprotect fuse programing */
	writeb(0xaa, IMX_IIM_BASE + IIM_PREG_P);

	/* upper half address register */
	writeb((bank << 3) | (row >> 7), IMX_IIM_BASE + IIM_UAF);
	printf("%s: upper half = 0x%02x\n", __func__,
			readb(IMX_IIM_BASE + IIM_UAF));

	for (bit = 0; bit < 8; bit++) {
		if (((value >> bit) & 1) == 0)
			continue;

		/* lower half address register */
		writeb(((row << 1) | bit), IMX_IIM_BASE + IIM_LA);
		printf("%s: lower half = 0x%02x\n", __func__,
				readb(IMX_IIM_BASE + IIM_LA));
		printf("%s: stat = 0x%02x\n", __func__,
				readb(IMX_IIM_BASE + IIM_STAT));

		/* start fuse programing */
		writeb(0x71, IMX_IIM_BASE + IIM_FCTL);

		/* wait for program done */
		while ((readb(IMX_IIM_BASE + IIM_STAT) & 0x80) != 0)
			;

		/* clear program done status */
		stat = readb(IMX_IIM_BASE + IIM_STAT);
		writeb(stat, IMX_IIM_BASE + IIM_STAT);
		printf("%s: stat = 0x%02x\n", __func__, stat);

		err = readb(IMX_IIM_BASE + IIM_ERR);
		if (err) {
			printf("fuse_blow: bit %d program error (0x%02x)\n",
					bit, err);
			ret = 1;
			goto out;
		}

		printf("%s: bit %d programed successfully\n", __func__, bit);
	}

out:
	/* protect fuse programing */
	writeb(0, IMX_IIM_BASE + IIM_PREG_P);
	return ret;
}

/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * U-Boot Command 
 *  bootstrap control
 */
int do_fuse_bootstrap(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int sdat;
	int ret = 1;

	if (argc != 2) {
		cmd_usage(cmdtp);
		return 1;
	}

	if (!memcmp(argv[1], "set", 3)) {
		printf("+++++++ setting fuse bootstrap ++++++++\n");
		/* Fuse blow
		 *  : BANK0 - 0x10 - GPIO_BT_SEL = 1
		 *  : BANK0 - 0x0C - BT_USB_SRC        = 0  (Internal UTMI PHY)
		 *  : BANK0 - 0x0C - BT_SPARE_SIZE     = 1  (218byte spare)
		 *  : BANK0 - 0x0C - BT_MLC_SEL        = 0  (SLC Device)
		 *  : BANK0 - 0x10 - BT_PAGE_SIZE[1:0] = 10 (4K bytes)
		 *  : BANK0 - 0x14 - BT_MEM_CTL[1:0]   = 01 (Nand Flash)
		 *  : BANK0 - 0x14 - BT_BUS_WIDTH[1:0] = 00 (8bit bus)
		 *  : BANK0 - 0x14 - BT_MEM_TYPE[1:0]  = 10 (5address cycles)
		 *  : BANK0 - 0x44 - BT_LPB_FREQ[2:0]  = 0  (133MHz)
		 */

		/* address : 000C
		 *  0000 0100 (04h)
		 */
		fuse_blow_value(0, 0x0C, 0x04);

		/* address : 0010
		 *  0100 1001 (49h)
		 */
		fuse_blow_value(0, 0x10, 0x49);

		/* address : 0014
		 *  0100 0010 (42h)
		 */
		fuse_blow_value(0, 0x14, 0x42);

		/* address : 0044
		 *  0000 0000
		 */
		/* Not Setting */
	} else if (!memcmp(argv[1], "disp", 4)) {
		printf("+++++++ display fuse sense ++++++++\n");
	} else {
		cmd_usage(cmdtp);
		return 1;
	}

	ret = 0;

	/* display fuse status */
	sdat = fuse_sence(0, 0x0C);
	if (sdat >= 0) {
		printf("bank 0 : row 0x0C: 0x%02x(%s)\n",
			(sdat & 0xff),  __convert_fuse(sdat & 0xff));
		printf(" - BT_USB_SRC     : %#x\n", (sdat & (1 << 0)) >> 0);
		printf(" - BT_SPARE_SIZE  : %#x\n", (sdat & (1 << 2)) >> 2);
		printf(" - BT_MLC_SEL     : %#x\n", (sdat & (1 << 5)) >> 5);
	}
	sdat = fuse_sence(0, 0x10);
	if (sdat >= 0) {
		printf("bank 0 : row 0x10: 0x%02x(%s)\n",
			(sdat & 0xff),  __convert_fuse(sdat & 0xff));
		printf(" - GPIO_BT_SEL    : %#x\n", (sdat & (1 << 3)) >> 3);
		printf(" - BT_PAGE_SIZE   : %#x\n", (sdat & (3 << 5)) >> 5);
	}
	sdat = fuse_sence(0, 0x14);
	if (sdat >= 0) {
		printf("bank 0 : row 0x14: 0x%02x(%s)\n",
			(sdat & 0xff),  __convert_fuse(sdat & 0xff));
		printf(" - BT_MEM_CTL     : %#x\n", (sdat & (3 << 1)) >> 1);
		printf(" - BT_BUS_WIDTH   : %#x\n", (sdat & (3 << 3)) >> 3);
		printf(" - BT_MEM_TYPE    : %#x\n", (sdat & (3 << 5)) >> 5);
	}
	sdat = fuse_sence(0, 0x44);
	if (sdat >= 0) {
		printf("bank 0 : row 0x14: 0x%02x(%s)\n",
			(sdat & 0x44),  __convert_fuse(sdat & 0xff));
		printf(" - BT_LPB_FREQ    : %#x\n", (sdat & (7 << 5)) >> 5);
	}

	return 0;
}
U_BOOT_CMD(
	fuse,	2,	1,	do_fuse_bootstrap,
	"Fuse BootStrap Setting command",
	": fuse <option> \n"
	"BootStrap fuses \"set\"  : setting bootstrap fuse\n"
	"                \"disp\" : disp bootstrap fuse\n"
);


