/*
 * (C) Copyright 2011
 * Working Solar System Production
 *  
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * UART Functions
 */

#include <common.h>
#include <asm/io.h>
#include <asm/errno.h>
#include <asm/arch/mx25.h>
#include <asm/arch/mx25-regs.h>
#include <asm/arch/mx25_pins.h>
#include <asm/arch/iomux.h>
#include <asm/arch/gpio.h>
#include <imx_spi.h>

//#define CMD_UART_DEBUG
#undef CMD_UART_DEBUG

#ifdef	CMD_UART_DEBUG
# define PRINTF(fmt,args...)	printf (fmt ,##args)
#else
# define PRINTF(fmt,args...)
#endif

/*-----------------------------------------------------------------------
 * UART Register
 */
#define __REG(x)	(*((volatile u32 *)(x)))

/* Register definitions */
#define URXD	0x0  /* Receiver Register */
#define UTXD	0x40 /* Transmitter Register */
#define UCR1	0x80 /* Control Register 1 */
#define UCR2	0x84 /* Control Register 2 */
#define UCR3	0x88 /* Control Register 3 */
#define UCR4	0x8c /* Control Register 4 */
#define UFCR	0x90 /* FIFO Control Register */
#define USR1	0x94 /* Status Register 1 */
#define USR2	0x98 /* Status Register 2 */
#define UESC	0x9c /* Escape Character Register */
#define UTIM	0xa0 /* Escape Timer Register */
#define UBIR	0xa4 /* BRM Incremental Register */
#define UBMR	0xa8 /* BRM Modulator Register */
#define UBRC	0xac /* Baud Rate Count Register */
#define UTS	0xb4 /* UART Test Register (mx25) */

/* UART Control Register Bit Fields.*/
#define URXD_CHARRDY	(1<<15)
#define URXD_ERR	(1<<14)
#define URXD_OVRRUN	(1<<13)
#define URXD_FRMERR	(1<<12)
#define URXD_BRK	(1<<11)
#define URXD_PRERR	(1<<10)
#define URXD_RX_DATA	(0xFF)
#define UCR1_ADEN	(1<<15)	/* Auto dectect interrupt */
#define UCR1_ADBR	(1<<14)	/* Auto detect baud rate */
#define UCR1_TRDYEN	(1<<13)	/* Transmitter ready interrupt enable */
#define UCR1_IDEN	(1<<12)	/* Idle condition interrupt */
#define UCR1_RRDYEN	(1<<9)	/* Recv ready interrupt enable */
#define UCR1_RDMAEN	(1<<8)	/* Recv ready DMA enable */
#define UCR1_IREN	(1<<7)	/* Infrared interface enable */
#define UCR1_TXMPTYEN	(1<<6)	/* Transimitter empty interrupt enable */
#define UCR1_RTSDEN	(1<<5)	/* RTS delta interrupt enable */
#define UCR1_SNDBRK	(1<<4)	/* Send break */
#define UCR1_TDMAEN	(1<<3)	/* Transmitter ready DMA enable */
#define UCR1_UARTCLKEN	(1<<2)	/* UART clock enabled */
#define UCR1_DOZE	(1<<1)	/* Doze */
#define UCR1_UARTEN	(1<<0)	/* UART enabled */
#define UCR2_ESCI	(1<<15)	/* Escape seq interrupt enable */
#define UCR2_IRTS	(1<<14)	/* Ignore RTS pin */
#define UCR2_CTSC	(1<<13)	/* CTS pin control */
#define UCR2_CTS	(1<<12)	/* Clear to send */
#define UCR2_ESCEN	(1<<11)	/* Escape enable */
#define UCR2_PREN	(1<<8)	/* Parity enable */
#define UCR2_PROE	(1<<7)	/* Parity odd/even */
#define UCR2_STPB	(1<<6)	/* Stop */
#define UCR2_WS		(1<<5)	/* Word size */
#define UCR2_RTSEN	(1<<4)	/* Request to send interrupt enable */
#define UCR2_TXEN	(1<<2)	/* Transmitter enabled */
#define UCR2_RXEN	(1<<1)	/* Receiver enabled */
#define UCR2_SRST	(1<<0)	/* SW reset */
#define UCR3_DTREN	(1<<13)	/* DTR interrupt enable */
#define UCR3_PARERREN	(1<<12)	/* Parity enable */
#define UCR3_FRAERREN	(1<<11)	/* Frame error interrupt enable */
#define UCR3_DSR	(1<<10)	/* Data set ready */
#define UCR3_DCD	(1<<9)	/* Data carrier detect */
#define UCR3_RI		(1<<8)	/* Ring indicator */
#define UCR3_TIMEOUTEN	(1<<7)	/* Timeout interrupt enable */
#define UCR3_RXDSEN	(1<<6)	/* Receive status interrupt enable */
#define UCR3_AIRINTEN	(1<<5)	/* Async IR wake interrupt enable */
#define UCR3_AWAKEN	(1<<4)	/* Async wake interrupt enable */
#define UCR3_REF25	(1<<3)	/* Ref freq 25 MHz */
#define UCR3_REF30	(1<<2)	/* Ref Freq 30 MHz */
#define UCR3_INVT	(1<<1)	/* Inverted Infrared transmission */
#define UCR3_BPEN	(1<<0)	/* Preset registers enable */
#define UCR4_CTSTL_32   (32<<10)/* CTS trigger level (32 chars) */
#define UCR4_INVR	(1<<9)	/* Inverted infrared reception */
#define UCR4_ENIRI	(1<<8)	/* Serial infrared interrupt enable */
#define UCR4_WKEN	(1<<7)	/* Wake interrupt enable */
#define UCR4_REF16	(1<<6)	/* Ref freq 16 MHz */
#define UCR4_IRSC	(1<<5)	/* IR special case */
#define UCR4_TCEN	(1<<3)	/* Transmit complete interrupt enable */
#define UCR4_BKEN	(1<<2)	/* Break condition interrupt enable */
#define UCR4_OREN	(1<<1)	/* Receiver overrun interrupt enable */
#define UCR4_DREN	(1<<0)	/* Recv data ready interrupt enable */
#define UFCR_RXTL_SHF	0	/* Receiver trigger level shift */
#define UFCR_RFDIV	(7<<7)	/* Reference freq divider mask */
#define UFCR_TXTL_SHF	10	/* Transmitter trigger level shift */
#define USR1_PARITYERR	(1<<15)	/* Parity error interrupt flag */
#define USR1_RTSS	(1<<14)	/* RTS pin status */
#define USR1_TRDY	(1<<13)	/* Transmitter ready interrupt/dma flag */
#define USR1_RTSD	(1<<12)	/* RTS delta */
#define USR1_ESCF	(1<<11)	/* Escape seq interrupt flag */
#define USR1_FRAMERR	(1<<10)	/* Frame error interrupt flag */
#define USR1_RRDY	(1<<9)	/* Receiver ready interrupt/dma flag */
#define USR1_TIMEOUT	(1<<7)	/* Receive timeout interrupt status */
#define USR1_RXDS	(1<<6)	/* Receiver idle interrupt flag */
#define USR1_AIRINT	(1<<5)	/* Async IR wake interrupt flag */
#define USR1_AWAKE	(1<<4)	/* Aysnc wake interrupt flag */
#define USR2_ADET	(1<<15)	/* Auto baud rate detect complete */
#define USR2_TXFE	(1<<14)	/* Transmit buffer FIFO empty */
#define USR2_DTRF	(1<<13)	/* DTR edge interrupt flag */
#define USR2_IDLE	(1<<12)	/* Idle condition */
#define USR2_IRINT	(1<<8)	/* Serial infrared interrupt flag */
#define USR2_WAKE	(1<<7)	/* Wake */
#define USR2_RTSF	(1<<4)	/* RTS edge interrupt flag */
#define USR2_TXDC	(1<<3)	/* Transmitter complete */
#define USR2_BRCD	(1<<2)	/* Break condition */
#define USR2_ORE	(1<<1)	/* Overrun error */
#define USR2_RDR	(1<<0)	/* Recv data ready */
#define UTS_FRCPERR	(1<<13)	/* Force parity error */
#define UTS_LOOP	(1<<12)	/* Loop tx and rx */
#define UTS_TXEMPTY	(1<<6)	/* TxFIFO empty */
#define UTS_RXEMPTY	(1<<5)	/* RxFIFO empty */
#define UTS_TXFULL	(1<<4)	/* TxFIFO full */
#define UTS_RXFULL	(1<<3)	/* RxFIFO full */
#define UTS_SOFTRST	(1<<0)	/* Software reset */

static void uart_set_baudrate(u32 base_reg, u32 baudrate)
{
	u32 clk = mx25_get_ipg_clk();

	__REG(base_reg + UFCR) = 4 << 7; /* divide input clock by 2 */
	__REG(base_reg + UBIR) = 0xf;
	__REG(base_reg + UBMR) = clk / (2 * baudrate);
}

/*-----------------------------------------------------------------------
 * UART1
 */
#define UART1_PHYS	UART1_BASE
#define MAX_CONSOLE_BUF	128
static char consolebuf[MAX_CONSOLE_BUF];

/*-----------------------------------------------------------------------
 * UART2
 */
#ifdef CONFIG_CMD_UART2_TEST

#define UART2_PHYS	UART2_BASE
#define UART2_BAUDRATE	2400
#define ECOCUTE_MAXBUF	8
#define RESPONSE_MAXBUF	5

/* Ecocute Data(CSI) */

/*
 * Command Format : 8byte
 * -------------------------------------------------
 * | CMD/ADR   | RECV/CODE | DATA1 - 5 | FCC       |
 * -------------------------------------------------
 */
#define CSI_CMD_ADR		(0)
#define CSI_RECV_CODE		(1)
#define CSI_DATA1		(2)
#define CSI_DATA2		(3)
#define CSI_DATA3		(4)
#define CSI_DATA4		(5)
#define CSI_DATA5		(6)
#define CSI_FCC			(7)

/* CMD/ADR */
#define CSI_CMD_MASK		(0xF0)
# define CSI_CMD_CODE_DRIVE	(0xA0)
# define CSI_CMD_CODE_REQ	(0x90)

#define CSI_ADR_MASK		(0x07)
# define CSI_ADR_CODE_MAIN	(0x01)
# define CSI_ADR_CODE_BATH	(0x02)
# define CSI_ADR_CODE_EXT	(0x03)

/* RECV/CODE */
#define CSI_RECV_MASK		(0x70)	/* bit 7 not used */
# define CSI_RECV_REMOCON_1	(0x10)
# define CSI_RECV_REMOCON_2	(0x20)
# define CSI_RECV_REMOCON_3	(0x40)

#define CSI_CODE_MASK		(0x0F)
# define CSI_CODE_0		(0x00)
# define CSI_CODE_1		(0x01)
# define CSI_CODE_2		(0x02)
# define CSI_CODE_3		(0x03)
# define CSI_CODE_4		(0x04)
# define CSI_CODE_5		(0x05)
# define CSI_CODE_6		(0x06)
# define CSI_CODE_7		(0x07)
# define CSI_CODE_8		(0x08)
# define CSI_CODE_9		(0x09)
# define CSI_CODE_A		(0x0A)

/* DATA1 - 5 */

/* Key Input */
static u8 uart2_input_key(void)
{
	static int key_pt = 0;

	if (serial_tstc()) {
		consolebuf[key_pt] = getc();
		switch (consolebuf[key_pt]) {
		case 0x0d:		/* '\n' */
			putc('\n');
			consolebuf[key_pt] = '\0';
			key_pt = 0;
			return (u8)(simple_strtoul(consolebuf, NULL, 16) & 0xff);
		case 0x03:		/* ^C - CTRL+C */
			putc('\n');
			key_pt = 0;
			return 0xfe;
		case 0x08:		/* '\b' */
			putc(consolebuf[key_pt]);
			if (key_pt >= 1) {
				key_pt--;
				consolebuf[key_pt] = ' ';
				putc(consolebuf[key_pt]);
				putc('\b');
			}
			break;
		default:
			putc(consolebuf[key_pt]);
			key_pt++;
			break;
		}
	}

	return 0xff;
}

/* FCC */
inline u8 uart2_csi_fcc (u8 *data, int len)
{
	u8 fcc = *data;
	int i;

	if (!len)
		return 0;

	for (i = 1; i < len; i++)
		fcc ^= *(data + i);

	return fcc;
}

/*
 * Initialise the uart2 port with the given baudrate. The settings
 * are always 8 data bits, no parity, 1 stop bit, no start bits.
 *
 */
static int uart2_init(void)
{
	static int uart2init = 0;

	if (uart2init)
		goto end;

	/* uart */
	__REG(UART2_PHYS + UCR1) = 0x0;
	__REG(UART2_PHYS + UCR2) = 0x0;

	while (!(__REG(UART2_PHYS + UCR2) & UCR2_SRST))
		;

	__REG(UART2_PHYS + UCR3) = 0x0704;
	__REG(UART2_PHYS + UCR4) = 0x8000;
	__REG(UART2_PHYS + UESC) = 0x002b;
	__REG(UART2_PHYS + UTIM) = 0x0;

	__REG(UART2_PHYS + UTS) = 0x0;

	uart_set_baudrate(UART2_PHYS, UART2_BAUDRATE);

	/* Parity Setting */
	__REG(UART2_PHYS + UCR2) = UCR2_WS | UCR2_IRTS | UCR2_RXEN |
				UCR2_TXEN | UCR2_SRST | UCR2_PREN ;

	__REG(UART2_PHYS + UCR1) = UCR1_UARTEN;

	uart2init = 1;
end:
	return 0;
}

/*
 * ecocute test program main
 *
 */
int do_uart2test (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	unsigned long empty;
	u8 eco[ECOCUTE_MAXBUF];
	u8 txdata[RESPONSE_MAXBUF];
	u8 txlen, rxlen;
	u8 data = 0;
	u8 temp;
	int i;

	uart2_init();
	printf("++++++++++++ Ecocute Serial Start ++++++++++++\n");
	printf(" - Test End is \"CTRL + C\"                   \n");
	printf(" - temperature Change is input Number         \n");
	printf("++++++++++++++++++++++++++++++++++++++++++++++\n");

	for (empty = 1; ;) {
		data = uart2_input_key();
		if (data == 0xfe) {	/* ctrl + c */
			return 0;
		} else if(data != 0xff) {
			empty = 0;
			temp = data;
		} else {
			/* Not input */
		}

		/* UART2 CSI Receive */
		if (__REG(UART2_PHYS + UTS) & UTS_RXEMPTY)
			continue;

		eco[CSI_CMD_ADR] = (u8)(__REG(UART2_PHYS + URXD) & URXD_RX_DATA);
		switch (eco[CSI_CMD_ADR] & CSI_CMD_MASK) {
		case CSI_CMD_CODE_DRIVE:
			rxlen = 8;
			break;
		case CSI_CMD_CODE_REQ:
			rxlen = 1;
			PRINTF("Eco(???) : %02X\n", eco[CSI_CMD_ADR]);
			break;
		default:
			PRINTF("Eco(???) : %02X\n", eco[CSI_CMD_ADR]);
			continue;
		}
		/* Not main */
		if ((eco[CSI_CMD_ADR] & CSI_ADR_MASK) != CSI_ADR_CODE_MAIN)
			continue;

		for (i = 1; i < rxlen; i++) {
			/* empty check */
			while (__REG(UART2_PHYS + UTS) & UTS_RXEMPTY)
				;
			/* read date */
			eco[i] = (u8)(__REG(UART2_PHYS + URXD) & URXD_RX_DATA);
		}

		if (rxlen == 8) {
			/* FCC check */
			if (eco[CSI_FCC] != uart2_csi_fcc(eco, 7))
				continue;
			printf("Eco(DRI) : %02X %02X %02X %02X %02X %02X %02X %02X\n",
				eco[0], eco[1], eco[2], eco[3], 
				eco[4], eco[5], eco[6], eco[7]);
		} else if (rxlen == 1) {
			printf("Eco(REQ) : %02X\n", eco[0]);
		}

		/* wait 4.6msec */
		udelay(4600);

		/* DSI or CQI return */
		if (empty) {
			txdata[0] = 0x61;	/* RP */
			txlen = 1;
		} else {
			txdata[0] = 0xB1;	/* DSI */
			txdata[1] = 0x32;
			txdata[2] = temp;
			txdata[3] = 0x00;
			txdata[4] = uart2_csi_fcc(txdata, 4);
			txlen = RESPONSE_MAXBUF;
			empty = 1;
		}
		for (i = 0; i < txlen; i++) {
			__REG(UART2_PHYS + UTXD) = (u8)txdata[i];
	
			/* wait for transmitter to be ready */
			while (!(__REG(UART2_PHYS + UTS) & UTS_TXEMPTY))
				;
		}

		if (txlen == 1)
			printf("Rmo(RP ) : %02X\n", txdata[0]);
		else if (txlen == 5)
			printf("Rmo(DSI) : %02X %02X %02X %02X %02X\n",
				txdata[0], txdata[1], txdata[2], txdata[3], txdata[4]);
	}
	return 0;
}

/* #define U_BOOT_CMD(name,maxargs,rep,cmd,usage,help) */
U_BOOT_CMD(
	uart2test,	1,	1,	do_uart2test,
	"UART2 test command",
	"uart2test .. end is \"CTRL + C\""
);

#endif	/* CONFIG_CMD_UART2_TEST */

/*-----------------------------------------------------------------------
 * UART3
 */
#ifdef CONFIG_CMD_UART3_TEST

#define UART3_PHYS	UART3_BASE
#define UART3_BAUDRATE	9600

/*+++++++++++++++++
 * Command
 *  remo ----> powcon
 */
const u8 send_com_powcon[] = {
	0x05,		/* ENQ */
	0x30, 0x36,	/* address */
	0x41,		/* parameter */
	0x41, 0x37,	/* checksum */
	0x03,		/* ETX */
	0x0D		/* CR */
};

/*+++++++++++++++++
 * Command
 *  remo ----> sensor
 */
const u8 send_com_sensor[] = {
	0x05,		/* ENQ */
	0x30, 0x31,	/* address */
	0x30,		/* version */
	0x30,		/* command */
	0x30,		/* parameter */
	0x46, 0x31,	/* checksum */
	0x03,		/* ETX */
	0x0D		/* CR */
};

/*
 * Initialise the uart2 port with the given baudrate. The settings
 * are always 8 data bits, no parity, 1 stop bit, no start bits.
 *
 */
static int uart3_init(void)
{
	static int uart3init = 0;

	if (uart3init)
		goto end;

	while (!(__REG(UART3_PHYS + UCR2) & UCR2_SRST))
		;

	__REG(UART3_PHYS + UCR3) = 0x0704;
	__REG(UART3_PHYS + UCR4) = 0x8000;
	__REG(UART3_PHYS + UESC) = 0x002b;
	__REG(UART3_PHYS + UTIM) = 0x0;

	__REG(UART3_PHYS + UTS) = 0x0;

	uart_set_baudrate(UART3_PHYS, UART3_BAUDRATE);

	__REG(UART3_PHYS + UCR2) = UCR2_WS | UCR2_IRTS | UCR2_RXEN |
					UCR2_TXEN | UCR2_SRST;

	__REG(UART3_PHYS + UCR1) = UCR1_UARTEN;

	uart3init = 1;
end:

	return 0;
}

/* uart3 test command main */
int do_uart3test (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
#define UART3_RX_TIMEOUT	100000
#define UART3_RX_TICK		1
	int len, i;
	int timeout;

	u8 temp;
	u8 res[128];
	u8 *command = send_com_powcon;	/* default : powercon */
	int txlen = 8;

	uart3_init();

	/* 
	 * Format
	 *  -------------------------------------------------------
	 * | EQN  |         |         |          |        |        |
	 * | or   | address | data    | checksum | ETX    | CR     |
	 * | STX  |         |         |          |        |        |
	 *  -------------------------------------------------------
	 */
	if (argc >= 2) {
		if (!memcmp(argv[1], "sensor", 6)) {
			command = send_com_sensor;
			txlen = 10;
		}
	}
	printf("Format  (%02d) : ", txlen);
	for (i = 0; i < txlen; i++) {
		printf("%02X ", *(command + i));
	}
	printf("\n");

	/* Command Send */
	mxc_set_gpio_dataout(MX25_PIN_KPP_COL2, 1);	/* nRE */
	mxc_set_gpio_dataout(MX25_PIN_CSI_MCLK, 1);	/* RS485 Enable */

	udelay(20 * 1000);

	for (i = 0; i < txlen; i++) {
		__REG(UART3_PHYS + UTXD) = *(command + i);

		/* wait for transmitter to be ready */
		while (!(__REG(UART3_PHYS + UTS) & UTS_TXEMPTY))
			;
	}
	for (timeout = 0; timeout < UART3_RX_TIMEOUT; timeout++) {
		if (__REG(UART3_PHYS + USR2) & USR2_TXDC) {
			/* RS485 Disable */
			udelay(20 * 1000);

			mxc_set_gpio_dataout(MX25_PIN_CSI_MCLK, 0);
			mxc_set_gpio_dataout(MX25_PIN_KPP_COL2, 0);	/* nRE */
			break;
		} else {
			udelay(1);
		}
	}

	/* Response Receive */
	for (len = 0; len < 128; len++) {
		/* empty check */
		for (timeout = 0; timeout < UART3_RX_TIMEOUT; timeout++) {
			if (!(__REG(UART3_PHYS + UTS) & UTS_RXEMPTY))
				break;
			else
				udelay(UART3_RX_TICK);
		}
		/* response timeout */
		if (timeout >= UART3_RX_TIMEOUT) {
			if (!len)
				printf("PowerCon timeout response...\n");
			break;
		}

		/* read data */
		res[len] = (u8)(__REG(UART3_PHYS + URXD) & URXD_RX_DATA);
	}
	printf("Response(%02d) : ", len);
	for (i = 0; i < len; i++) {
		printf("%02X ", res[i]);
	}
	printf("\n");

	return 0;
failed:
	return 1;
}

/* #define U_BOOT_CMD(name,maxargs,rep,cmd,usage,help) */
U_BOOT_CMD(
	uart3test,	2,	1,	do_uart3test,
	"UART3 test command",
	" power  -- remo - powcon\n"
	" sensor -- remo - sensor\n"
);

#endif	/* CONFIG_CMD_UART3_TEST */

